package texture

import "../../tlr"
import "../../tlr/gpu"
import "core:image/qoi"
import "core:log"

Vertex :: struct{
	position: [2]f32,
	uv: [2]f32,
}

vertices := []Vertex{
	{{ 0.5, 0.5 }, { 1, 1 }},
	{{ 0.5, -0.5 }, { 1, 0 }},
	{{ -0.5, 0.5 }, { 0, 1 }},
	{{ -0.5, -0.5 }, { 0, 0 }},
}

indicies := []u32{
	0, 1, 2,
	1, 3, 2,
}

main :: proc() {
	context.logger = log.create_console_logger()
	defer log.destroy_console_logger(context.logger)

	when gpu.BACKEND == .Vulkan {
		gpu.vk_set_hints(gpu.Vk_Hints{
			use_validation_layers = true,
		})
	}

	tlr.init("Texture Example", 500, 500)
	gpu.init()
	defer {
		gpu.cleanup()
		tlr.cleanup()
	}

	vert_shader, vert_ok := gpu.make_shader("./examples/texture/texture.vert.spv")
	frag_shader, frag_ok := gpu.make_shader("./examples/texture/texture.frag.spv")
	assert(vert_ok && frag_ok)

	dice_image, dice_image_err := qoi.load_from_file("./examples/texture/dice.qoi")
	assert(dice_image_err == nil)
	defer qoi.destroy(dice_image)

	texture, _ := gpu.make_texture(gpu.Texture_Info{
		writable = false,
		size = [2]int{ dice_image.width, dice_image.height },
	})
	gpu.upload_texture(texture, dice_image.pixels.buf[:])

	descriptor_info := gpu.Descriptor_Info{
		binding = 0,
		count = 1,
		type = .Texture,
		stages = { .Fragment },
	}
	descriptors, _ := gpu.make_descriptors({ descriptor_info })
	gpu.update_descriptors_texture(descriptors, 0, texture)

	vertex_binding := gpu.Vertex_Binding{
		binding = 0,
		stride = size_of(Vertex),
		attributes = {
			{
				location = 0,
				format = .Vec2_Float32,
				offset = offset_of(Vertex, position),
			},
			{
				location = 1,
				format = .Vec2_Float32,
				offset = offset_of(Vertex, uv),
			},
		},
	}
	pipeline_info := gpu.Pipeline_Info{
		vertex = vert_shader,
		fragment = frag_shader,

		vertex_bindings = { vertex_binding, },
		descriptors = { descriptors },
	}
	pipeline, _ := gpu.make_pipeline(pipeline_info)

	vert_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(Vertex)*len(vertices),
	})
	gpu.upload_buffer(vert_buffer, vertices)

	ind_buffer, _ := gpu.make_buffer(gpu.Buffer_Info{
		size = size_of(u32)*len(indicies),
	})
	gpu.upload_buffer(ind_buffer, indicies)

	for !tlr.window_should_close() {
		using gpu
		begin_pass(Pass_Info{
			clear_color = [4]f32{ 0, 0.5, 0.5, 1 },
		})
		bind_pipeline(pipeline)
		bind_vertex_buffer(vert_buffer)
		bind_index_buffer(ind_buffer, .UInt32)
		bind_descriptors(descriptors)
		draw_elems_indexed(len(indicies))
		end_pass()
	}
}
