#!/usr/bin/python3
import re
from pathlib import Path
from tokenize import tokenize
import os

FILE_HEADER = """
// Auto generated bindings from the xcb headers
package {}

"""[1::]
CUSTOM_STRUCTS = """
iterator_t :: struct($T: typeid){
	data: [^]T,
	rem: i32,
	index: i32,
}

iovec :: struct{
	iov_base: rawptr,
	iov_len: int,
}

"""

src_core = ''
src_icccm = ''
for child in Path('./').iterdir():
	extension = child.suffix
	if child.is_file() and extension == '.h':
		if child.name.find('icccm') == -1:
			src_core += child.read_text() + '\n'
		else:
			src_icccm += child.read_text() + '\n'

prefixes = ['xcb_icccm_', 'xcb_']
prefixes_reg = '(xcb_icccm_|xcb_)'

def remove_prefix(text, prefix):
	if text.startswith(prefix):
		return text[len(prefix):]
	return text

def remove_prefixes(text, prefixes):
	for prefix in prefixes:
		if text.startswith(prefix):
			return text[len(prefix):]
	return text

def remove_suffix(text, suffix):
	if text.endswith(suffix):
		return text[:-len(suffix)]
	return text

def remove_suffixes(text, suffixes):
	for suffix in suffixes:
		if text.endswith(suffix):
			return text[:-len(suffix)]
	return text

iterators = []
def parse_type(t, import_types):
	t = remove_prefix(t, 'const')
	t = remove_prefix(t, 'extern').strip()
	table = {
		"xcb_bool32_t": 'b32',
		"float": 'f32',
		"double": 'f64',
		'uint8_t': 'u8',
		"uint16_t": 'u16',
		"uint32_t": 'u32',
		"uint64_t": 'u64',
		"size_t": 'int',
		'int': 'ctypes.int',
		'unsigned int': 'ctypes.uint',
		'int8_t': 'i8',
		"int16_t": 'i16',
		'int32_t': 'i32',
		'int64_t': 'i64',
		"char": "byte",
		"void": "void",
		"void*": "rawptr",
		"void *": "rawptr",
		"char*": 'cstring',
		"char *": 'cstring',
		'iovec': 'iovec',
	}

	if t in table.keys():
		return table[t]

	if t == '':
		return t
	elif t.endswith("*"):
		ttype = t[:len(t)-1]
		return "^{}".format(parse_type(ttype, import_types))
	elif t.startswith('xcb_'):
		t = remove_prefixes(t, prefixes)
		if t.endswith('_iterator_t'):
			for iter_name, iter_type in iterators:
				if iter_name == t:
					t = 'iterator_t({})'.format(iter_type)
		else:
			found = False
			for imp in import_types:
				for type_ in imp[0]:
					if type_ == t:
						found = True
						break

				if found:
					t = imp[2] + '.' + t
					break
		return t

	print(t)

def parse_array(n, t, import_types):
	name, length = n.split('[', 1)
	length = length[:-1]
	type_ = "[{}]{}".format(length, parse_type(t, import_types))
	return name, type_

def parse_fields(f, fields, import_types):
	fields = re.findall(r"\s+(.+?)\s+([*]?)([_a-zA-Z0-9[\]]+)(, [_a-zA-Z0-9[\]]+)?;", fields)
	ffields = []
	for type_, star, fname, additional in fields:
		t = ''
		type_ += star
		if '[' in fname:
			fname, t = parse_array(fname, type_, import_types)
		else:
			t = parse_type(type_, import_types)
		
		if fname in ['context', 'map']:
			fname += '_'
		ffields.append(tuple([fname, t]))
		
		if additional != '':
			additional = remove_prefix(additional, ', ')
			ffields.append(tuple([additional, t]))
	
	max_len = max([len(n) for n, _ in ffields], default=0)

	for n, t in ffields:
		k = max_len-len(n)+len(t)
		f.write("\t{}: {},\n".format(n, t.rjust(k)))

event_types = []
def add_event_type (et):
	excludes = ['randr_', 'xevie_', 'xkb_', 'xfixes_', 'glx_', 'x_print_', 'input_', 'xv_', 'sync_', 'screensaver_', 'shm_', 'shape_', 'present_', 'damage_']
	for exclude in excludes:
		if et.startswith(exclude):
			return
	event_types.append(remove_suffix(et, '_event_t'))

def parse_structs(f, src, collect_types, import_types):
	# Opaque handles
	data = re.findall(r"typedef struct {}(\w+?) {}(\w+?);".format(prefixes_reg, prefixes_reg), src, re.S)
	for _, _, _, name1 in data:
		f.write('{} :: struct {{}}\n'.format(name1))
		collect_types.append(name1)
	f.write('\n')

	data = re.findall(r"typedef struct {{(.+?)}} {}(\w+);".format(prefixes_reg), src, re.S)
	for fields, _, name in data:
		f.write("{} :: struct{{\n".format(name))
		parse_fields(f, fields, import_types)
		f.write('}\n\n')
		collect_types.append(name)

	data = re.findall(r"typedef (struct|union) {}(\w+) {{(.+?)}} (\w+|\w+ \w+)?;".format(prefixes_reg), src, re.S)

	for data_type, _, name, fields, _ in data:
		if name.endswith('iterator_t'):
			data_field = remove_prefixes(re.findall(r"(\w+) \*data;", fields, re.S)[0], prefixes)
			iterators.append(tuple([name, data_field]))
			continue
		elif name.endswith('event_t'):
			add_event_type(name)

		f.write("{} :: struct".format(name))
		if data_type == "union":
			f.write("#raw_union ")
		f.write("{\n")
		parse_fields(f, fields, import_types)
		f.write("}\n\n")
		collect_types.append(name)

def parse_enum_fields(f, fname, prefix, fields):
	fields = re.findall(r"\s+(.+?) = [(]?([1L]+ << [0-9]+|[0-9]+)[)]?,?", fields)

	for name, number in fields:
		name = remove_prefix(name, '{}{}_'.format(prefix.upper(), remove_suffixes(remove_suffix(fname, '_t'), ['_enum', 's_flags']).upper()))
		if name[0].isdigit():
			name = '_' + name
		f.write('\t{} = {},\n'.format(name, number.replace('L', '')))

def parse_enums(f, src, collect_types):
	data = re.findall(r"typedef enum {}(\w+?) {{(.+?)}} \w+?;".format(prefixes_reg), src, re.S)

	for prefix, name, fields in data:
		f.write("{} :: enum ctypes.uint {{\n".format(name))
		parse_enum_fields(f, name, prefix, fields)
		f.write("}\n\n")
		collect_types.append(name)

	data = re.findall(r"typedef enum {{(.+?)}} {}(\w+);".format(prefixes_reg), src, re.S)

	for fields, prefix, name in data:
		f.write("{} :: enum ctypes.uint {{\n".format(name))
		parse_enum_fields(f, name, prefix, fields)
		f.write("}\n\n")
		collect_types.append(name)

def parse_parameters(parameters, import_types):
	params = []
	param = re.findall(r"(const |struct )?(unsigned )?(\w+)\s+([\*]+)?([_a-zA-Z0-9[\]]+)(  /\*\*< \*/)?", parameters)
	for _, unsigned, type_, star, pname, _ in param:
		t = ''
		type_ = unsigned + type_ + star
		if '[' in pname:
			pname, t = parse_array(pname, type_, import_types)
		else:
			t = parse_type(type_, import_types)

		if pname in ['context', 'map']:
			pname += '_'
		params.append(tuple([pname, t]))

	return params

def parse_bindings(f, src, prefix, libname, import_types):
	f.write("""
foreign import libxcb "system:{}"
@(default_calling_convention = "c", link_prefix = "{}")
foreign libxcb {{
""".format(libname, prefix))

	data = re.findall(r"^(const struct |const |struct )?(\w+)(\*| \*)?[\s]?{}(\w+)\s?\((.+?)\);".format(prefix), src, re.MULTILINE|re.S)
	for _, ret_type, star, name, params in data:
		ret_type += star
		params = params.split(',')

		pparams = []
		for param in params:
			param_str = param.strip()
			proc_param = re.findall(r"(\w+) \(\*(\w+)\)\((.+)\)", param_str)
			if len(proc_param) > 0:
				for arg_ret_type, arg_name, arg_params in proc_param:
					proc_params = parse_parameters(arg_params, import_types)
					pt = ''
					for proc_param_name, proc_param_type in proc_params:
						pt += '{}: {}, '.format(proc_param_name, proc_param_type)
					t = '#type proc "c" ({})'.format(pt)
					arg_ret_type = parse_type(arg_ret_type, import_types)
					if arg_ret_type != 'void':
						t += ' -> {}'.format(arg_ret_type)
					pparams.append(tuple([arg_name, t]))
			else:
				pparams += parse_parameters(param_str, import_types)
		
		f.write('\t{} :: proc('.format(name))
		for n, t in pparams:
			f.write('{}: {}, '.format(n, t))
		ret_type = parse_type(ret_type, import_types)
		if ret_type == 'void':
			f.write(') ---\n')
		else:
			f.write(') -> {} ---\n'.format(ret_type))
	f.write('}\n')

def parse_constants(f, src):
	data = re.findall(r"#define {}(\w+) (\w+)".format(prefixes_reg.upper()), src, re.S)
	for _, name, value in data:
		if not value[0].isdigit():
			continue

		value = remove_suffix(value, 'L')
		f.write('{} :: {}\n'.format(name, value))

def parse_aliases(f, src, collect_types):
	data = re.findall(r"typedef (\w+) {}(\w+);".format(prefixes_reg), src, re.S)
	for type_, _, name in data:
		if name.endswith('event_t'):
			f.write('{} :: distinct {}\n'.format(name, parse_type(type_, [])))
			add_event_type(name)
		else:
			f.write('{} :: {}\n'.format(name, parse_type(type_, [])))

		collect_types.append(name)


def gen_bindings(package_name, src, root, prefix, libname, custom_structs, import_types):
	root_folder = Path(root)
	if not root_folder.exists():
		os.makedirs(root_folder)

	imp_statements = 'import ctypes \"core:c\"\n'
	for imp in import_types:
		imp_statements += 'import {} "{}"\n'.format(imp[2], imp[1])

	types = []
	with open(root + 'structs.odin', 'w', encoding='utf-8') as f:
		f.write(FILE_HEADER.format(package_name))
		f.write(imp_statements)
		f.write(custom_structs)
		parse_structs(f, src, types, import_types)

	with open(root + 'enums.odin', 'w', encoding='utf-8') as f:
		f.write(FILE_HEADER.format(package_name))
		f.write(imp_statements)
		parse_enums(f, src, types)

	with open(root + 'bindings.odin', 'w', encoding='utf-8') as f:
		f.write(FILE_HEADER.format(package_name))
		f.write(imp_statements)
		parse_bindings(f, src, prefix, libname, import_types)

	with open(root + 'constants.odin', 'w', encoding='utf-8') as f:
		f.write(FILE_HEADER.format(package_name))
		f.write(imp_statements)
		parse_aliases(f, src, types)
		parse_constants(f, src)
	
	return types

types = gen_bindings('xcb', src_core, '../', 'xcb_', 'xcb', CUSTOM_STRUCTS, [])
gen_bindings('icccm', src_icccm, '../icccm/', 'xcb_icccm_', 'xcb-icccm', '', [tuple([types, '..', 'xcb'])])

with open('../utils.odin', 'w', encoding='utf-8') as f:
	UTILS_HEADER = """
package xcb

import ctypes "core:c"

EVENT_RESPONSE_TYPE_MASK_INV :: ~u8(0x80)

event_response_type :: #force_inline proc(response_type: u8) -> u8 {
	return response_type & EVENT_RESPONSE_TYPE_MASK_INV
}
"""[1::]
	f.write(UTILS_HEADER)
	f.write('\nEvent :: union {\n')
	for event_type in event_types:
		f.write('\t^{}_event_t,\n'.format(event_type))
	f.write('}\n\n')
	
	ITER_EVENTS_SRC = """
iter_events :: proc(c: ^connection_t) -> (e: Event, has_next: bool) {{
	generic_event := poll_for_event(c)
	if generic_event == nil {{
		has_next = false
		return
	}}

	switch event_response_type(generic_event.response_type) {{
{}	case: e = nil
	}}
	has_next = true
	return
}}
"""
	cases = ''
	for event_type in event_types:
		cases += '\tcase {}: e = transmute(^{}_event_t)generic_event\n'.format(event_type.upper(), event_type)

	f.write(ITER_EVENTS_SRC.format(cases))

